#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif
#else
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#ifdef WIN32
#define USE_OLDSIGNAL
#endif

#define BUFFER_SIZE (1024*80)

#define LOG(args...) do { fprintf(stderr, args); } while (0)
#define VLOG(args...) do { if (verbose) fprintf(stderr, args); } while (0)
#define ERROR_EXIT(args...) do { fprintf(stderr, args); exit(1); } while (0)
#define GOTO_DONE(args...) do { fprintf(stderr, args); goto done; } while (0)

static void help()
{
  printf("tcp-request [-v] <hostname> <port>\n");
  exit(0);
}

static int terminated = 0;
static void sigint_handler(int value)
{
  terminated = 1;
}

#ifdef WIN32
int winsock_init()
{
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    ERROR_EXIT("WSAStartup() failed.\n");
  return 0;
}
#endif

int main(int argc, char *argv[])
{
  char *hostname;
  int verbose = 0, port, s, readsize, sendsize, recvsize, writesize;
  int total_readsize = 0, total_sendsize = 0;
  int total_recvsize = 0, total_writesize = 0;
  struct sockaddr_in addr;
  socklen_t addrlen;
  char buffer[BUFFER_SIZE];
#ifdef USE_GETHOSTBYNAME
  struct hostent *host;
#endif
#ifndef USE_OLDSIGNAL
  struct sigaction sa;
#endif

  argc--; argv++;

  if ((argc > 0) && !strcmp(argv[0], "-v")) {
    verbose = 1;
    argc--; argv++;
  }

  if (argc < 2)
    help();

  hostname = *argv;
  argc--; argv++;
  port = atoi(*argv);
  argc--; argv++;

#ifdef WIN32
  winsock_init();
  if (!isatty(0))
    setmode(0, O_BINARY);
  if (!isatty(1))
    setmode(1, O_BINARY);
#endif

#ifdef USE_GETHOSTBYNAME
  /* ۥ̾饢ɥ쥹 */
  host = gethostbyname(hostname);
  if (host == NULL)
    ERROR_EXIT("Cannot resolve hostname.\n");
#endif

  /* åȤΥץ */
  s = socket(PF_INET, SOCK_STREAM, 0);
  if (s < 0)
    ERROR_EXIT("Cannot open socket.\n");

  /* ɥ쥹 */
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
#ifdef USE_GETHOSTBYNAME
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);
#else
  addr.sin_addr.s_addr = inet_addr(hostname);
#endif
  addrlen = sizeof(addr);

  /* Ф³ */
  if (connect(s, (struct sockaddr *)&addr, addrlen) < 0)
    ERROR_EXIT("Cannot connect.\n");

  /* ʥϥɥϿ */
#ifndef USE_OLDSIGNAL
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;
  if (sigaction(SIGINT, &sa, NULL) < 0)
#else
  if (signal(SIGINT, sigint_handler) == SIG_ERR)
#endif
    ERROR_EXIT("Cannot set signal.\n");

  while (!terminated) {
    /* ɸϤǡɤ߹ */
#ifndef WIN32
    readsize = read(0, buffer, sizeof(buffer));
#else
    /*
     * WindowsɸϤ(ѥפǤʤ)󥽡ξ硤read()Ϳ
     * Хåե礭read()Ԥƥ顼֤Τǡ
     * ХåեĴ롥
     */
#define MAXSIZE (1024*4)
#define ADJSIZE(size) ((isatty(0) && ((size) > MAXSIZE)) ? MAXSIZE : (size))
    readsize = read(0, buffer, ADJSIZE(sizeof(buffer)));
#endif
    if (readsize < 0)
      GOTO_DONE("Cannot read.\n");
    if (readsize == 0)
      break;

    VLOG("Read size from stdin:\t%d\n", readsize);
    total_readsize += readsize;

    /* ǡ */
#ifndef WIN32
    sendsize = write(s, buffer, readsize);
#else
    sendsize = send(s, buffer, readsize, 0);
#endif
    if (sendsize < 0)
      GOTO_DONE("Cannot send.\n");

    VLOG("Send size on TCP:\t%d\n", sendsize);
    total_sendsize += sendsize;

    /* ǡμ */
#ifndef WIN32
    recvsize = read(s, buffer, sizeof(buffer));
#else
    recvsize = recv(s, buffer, sizeof(buffer), 0);
#endif
    if (recvsize < 0)
      GOTO_DONE("Cannot recv.\n");
    if (recvsize == 0)
      break;

    VLOG("Recv size on TCP:\t%d\n", recvsize);
    total_recvsize += recvsize;

    /* ɸϤ˥ǡ񤭽Ф */
    writesize = write(1, buffer, recvsize);
    if (writesize < 0)
      GOTO_DONE("Cannot write.\n");
    VLOG("\n");

    VLOG("Write size to stdout:\t%d\n", writesize);
    total_writesize += writesize;
  }

done:
  if (terminated)
    LOG("Terminated.\n");

  VLOG("\n");
  VLOG("Total read size:\t%d\n", total_readsize);
  VLOG("Total send size:\t%d\n", total_sendsize);
  VLOG("Total recv size:\t%d\n", total_recvsize);
  VLOG("Total write size:\t%d\n", total_writesize);

  /* åȤΥ */
#ifndef WIN32
  shutdown(s, SHUT_RDWR);
  close(s);
#else
  shutdown(s, SD_BOTH);
  closesocket(s);
#endif

  return 0;
}
